//  Histogram and Image Statistics Library
//
//  efg, April 1998

unit gUHistogramLibrary;

interface

uses
  Windows, // TRGBTriple
  Graphics, // TCanvas
  gUStatisticsLibrary, // TDescriptiveStatitics
  gUColorLibrary; // TColorPlane

type
  THistoArray = array[BYTE] of INTEGER;

  TRGBHistoArray =
    record
    Red: THistoArray;
    Green: THistoArray;
    Blue: THistoArray;
    Intensity: THistoArray;
  end;

  THistogram =
    class(TObject)
  private
    FHistogram: THistoArray;
    function GetCount: INTEGER;

  public
    constructor Create;

    procedure Clear;
    procedure Increment(const index: BYTE);
    procedure GetStatistics(var N: INTEGER;
      var Minimum, Maximum: BYTE;
      var Mode, Median: BYTE;
      var Mean, StandardDeviation,
      Skewness, Kurtosis: DOUBLE);
    property Frequency: THistoArray read FHistogram write FHistogram;
    property Count: INTEGER read GetCount;
  end;

  TRGBHistograms =
    class
  protected
    FRed: THistogram;
    FGreen: THistogram;
    FBlue: THistogram;
    FIntensity: THistogram;

  public
    constructor Create;
    destructor Destroy; override;

    procedure Increment(const RGBTriple: TRGBTriple);

    property Red: THistogram read FRed;
    property Green: THistogram read FGreen;
    property Blue: THistogram read FBlue;
  end;

  TRGBStatistics =
    class(TObject)
  protected
    FRed: TDescriptiveStatistics;
    FGreen: TDescriptiveStatistics;
    FBlue: TDescriptiveStatistics;

    function PixelCount: INTEGER;

  public
    property Count: INTEGER read PixelCount;
    property Red: TDescriptiveStatistics read FRed;
    property Green: TDescriptiveStatistics read FGreen;
    property Blue: TDescriptiveStatistics read FBlue;

    constructor Create;
    destructor Destroy; override;
    procedure NextRGBTriple(const rgb: TRGBTriple);
    procedure ProcessBitmap(const Bitmap: TBitmap);
    procedure ResetValues;
  end;

procedure DrawHistogram(const ColorPlane: TColorPlane;
  const Histogram: THistogram;
  const Canvas: TCanvas);
procedure GetHistogram(const Bitmap: TBitmap;
  const ColorPlane: TColorPlane;
  var Histogram: THistogram);
function GetRGBHistograms(const Bitmap: TBitmap;
  const Rect: TRect): TRGBHistograms;


  // Get statistics for single image plane.  Use TRGBStatistics.ProcessBitmap
  // to get R, G, and B statistics for any number of 24-bit images.
procedure GetBitmapStatistics(const Bitmap: TBitmap;
  const ColorPlane: TColorPlane;
  var Statistics: TDescriptiveStatistics);

implementation

uses
  Dialogs, // ShowMessage
  gGlobal,
  gUImageProcessingPrimitives, // RGBTriple
  Math, // MaxIntValue, IntPwr
  SysUtils; // pByteArray, Exception

type
  EHistogramError = class(Exception);
  EStatisticsError = class(Exception);


  // == THistogram ======================================================

  // This Histogram class is specialized for image processing applications.
  // The frequency distribution is assumed to be for values ONLY in the
  // range 0..255.


function THistogram.GetCount: INTEGER;
var
  i: BYTE;
begin
  RESULT := 0;
  for i := Low(THistoArray) to High(THistoArray) do
    INC(RESULT, FHistogram[i])
end {GetCount};


constructor THistogram.Create;
begin
  Clear
end {Create};


procedure THistogram.Clear;
var
  i: BYTE;
begin
  for i := Low(THistoArray) to High(THistoArray) do
    FHistogram[i] := 0
end {Clear};


procedure THistogram.GetStatistics(var N: INTEGER;
  var Minimum, Maximum: BYTE;
  var Mode, Median: BYTE;
  var Mean, StandardDeviation,
  Skewness, Kurtosis: DOUBLE);
var
  Cumulative: INTEGER;
  i: BYTE;
  MaxFrequency: INTEGER;
  M2: EXTENDED;
  M3: EXTENDED;
  M3Sum: EXTENDED;
  M4: EXTENDED;
  M4Sum: EXTENDED;
  x: EXTENDED;
  xSum: EXTENDED; // Use floats to avoid integer overflow
  xSqrSum: EXTENDED;
begin
  N := 0;

  Minimum := 0;
  while (FHistogram[Minimum] = 0) and (Minimum < 255)
    do INC(Minimum);

  Maximum := 255;
  while (FHistogram[Maximum] = 0) and (Maximum > 0)
    do DEC(Maximum);

    // Mode is value with highest frequency.
    // For now, don't worry about a "tie".
  Mode := Minimum;
  MaxFrequency := FHistogram[Minimum];
  for i := Minimum + 1 to Maximum do
  begin
    if FHistogram[i] > MaxFrequency
      then begin
      Mode := i;
      MaxFrequency := FHistogram[i]
    end
  end;

    // Calculate Mean and Standard Deviation
  xSum := 0.0;
  xSqrSum := 0.0;
  for i := Minimum to Maximum do
  begin
    INC(N, FHistogram[i]);
    x := i;
    xSum := xSum + FHistogram[i] * x;
    xSqrSum := xSqrSum + FHistogram[i] * SQR(x)
  end;

  if N = 0
    then Mean := 0.0
  else Mean := xSum / N;

  if N < 2
    then begin
    StandardDeviation := 0.0; // should be a NAN someday
    Skewness := 0.0;
    Kurtosis := 0.0;
  end
  else begin
    StandardDeviation := SQRT((xSqrSum - N * SQR(Mean)) / (N - 1));

      // Standard Deviation is related to moment M2
    M2 := SQR(StandardDeviation) * (N - 1) / N;

      // Calculate third and fourth moments
    M3Sum := 0.0;
    M4Sum := 0.0;
    for i := Minimum to Maximum do
    begin
      x := i;

      M3Sum := M3Sum + FHistogram[i] * IntPower(x - Mean, 3);
      M4Sum := M4Sum + FHistogram[i] * IntPower(x - Mean, 4);
    end;

    M3 := M3Sum / N;
    M4 := M4Sum / N;

    if M2 = 0.0
      then begin
      Skewness := 0.0; // eventually use NAN here
      Kurtosis := 0.0;
    end
    else begin
      Skewness := M3 / Power(M2, 1.5);
      Kurtosis := M4 / SQR(M2)
    end
  end;

    // Median is value with half of values above and below.
  Cumulative := 0;
  i := Minimum;
  while (Cumulative < N div 2) and (i < 255) do
  begin
    INC(Cumulative, FHistogram[i]);
    INC(i)
  end;
  Median := i;

end {GetStatistics};


procedure THistogram.Increment(const index: BYTE);
begin
  INC(FHistogram[index])
end {Increment};


  // ==  RGB Histograms =================================================

constructor TRGBHistograms.Create;
begin
  FRed := THistogram.Create;
  FGreen := THistogram.Create;
  FBlue := THistogram.Create;
  FIntensity := THistogram.Create
end {Create};


destructor TRGBHistograms.Destroy;
begin
  FRed.Free;
  FGreen.Free;
  FBlue.Free;
  FIntensity.Free
end {Destroy};


procedure TRGBHistograms.Increment(const RGBTriple: TRGBTriple);
begin
  with RGBTriple do
  begin
    FRed.Increment(rgbtRed);
    FGreen.Increment(rgbtGreen);
    FBlue.Increment(rgbtBlue)
  end;
  FIntensity.Increment(RGBTripleToIntensity(RGBTriple));
end {Increment};




  // ==  RGB Statistics =================================================

constructor TRGBStatistics.Create;
begin
  FRed := TDescriptiveStatistics.Create;
  FGreen := TDescriptiveStatistics.Create;
  FBlue := TDescriptiveStatistics.Create;

  FRed.ResetValues;
  FGreen.ResetValues;
  FBlue.ResetValues
end {Create};


destructor TRGBStatistics.Destroy;
begin
  FRed.Free;
  FGreen.Free;
  FBlue.Free
end {Destroy};


function TRGBStatistics.PixelCount;
begin
  RESULT := FRed.Count
end {PixelCount};


  // Use this method to look at given set of pixels, one-by-one

procedure TRGBStatistics.NextRGBTriple(const rgb: TRGBTriple);
begin
  with rgb do
  begin
    FRed.NextValue(rgbtRed);
    FGreen.NextValue(rgbtGreen);
    FBlue.NextValue(rgbtBlue);
  end
end {NextRGBTriple};


  // Get statistics for complete image

procedure TRGBStatistics.ProcessBitmap(const Bitmap: TBitmap);
var
  i: INTEGER;
  j: INTEGER;
  row: pRGBTripleArray;
begin
  if Bitmap.PixelFormat <> pf24bit
    then
    raise EStatisticsError.Create('Can only process 24-bit image');

  for j := 0 to Bitmap.Height - 1 do
  begin
    row := Bitmap.Scanline[j];
    for i := 0 to Bitmap.Width - 1 do
    begin
      with row[i] do
      begin
        FRed.NextValue(rgbtRed);
        FGreen.NextValue(rgbtGreen);
        FBlue.NextValue(rgbtBlue);
      end
    end
  end
end {ProcessBitmap};


procedure TRGBStatistics.ResetValues;
begin
  FRed.ResetValues;
  FGreen.ResetValues;
  FBlue.ResetValues
end {ResetValues};


  // =====================================================================

procedure DrawHistogram(const ColorPlane: TColorPlane;
  const Histogram: THistogram;
  const Canvas: TCanvas);
const
  MajorTickSize = 8;

var
  BarLength: INTEGER;
  Delta: INTEGER;
  Color: TColor;
  i: INTEGER;
  j: INTEGER;
  Height: INTEGER;
  MaxValue: INTEGER;
  Width: INTEGER;
begin
  Color := clBlack; {avoid compiler warning about initialization}

  Height := Canvas.ClipRect.Bottom;
  Width := Canvas.ClipRect.Right;

  MaxValue := MaxIntValue(Histogram.Frequency);

    // For now only paint on a canvas exactly 256 pixels wide.  If
    // MaxValue is zero, array was not filled in correctly and is ignored.
  if (Width = 256) and (MaxValue > 0)
    then begin
    for i := Low(THistoArray) to High(THistoArray) do
    begin
      case ColorPlane of
        cpRGB,
          cpY,
          cpIntensity: Color := RGB(i, i, i);

        cpRed: Color := RGB(i, 0, 0);
        cpGreen: Color := RGB(0, i, 0);
        cpBlue: Color := RGB(0, 0, i);
        cpHueHSV:
              // Rescale Hue from 0..255 to 0..359
          Color := RGBTripleToColor(HSVtoRGBTriple(MulDiv(i, 359,
            255), 255, 255));

        cpSaturationHSV:
          Color := clBlack;
        cpValue: Color := RGB(i, i, i);

        cpHueHLS:
              // Rescale Hue from 0..255 to 0..359;  Half Lightness
          Color := RGBTripleToColor(HLStoRGBTriple(MulDiv(i, 359,
            255), 128, 255));

        cpSaturationHLS:
          Color := clBlack;

        cpLightness: Color := RGB(i, i, i);

        cpCyan: Color := RGB(i, i, i);
        cpMagenta: Color := RGB(i, i, i);
        cpYellow: Color := RGB(i, i, i);
        cpBlack: Color := RGB(255 - i, 255 - i, 255 - i)

      end;

      Canvas.Pen.Color := Color;
      BarLength := ROUND(Height * Histogram.Frequency[i] / MaxValue);
      Canvas.MoveTo(i, Height - 1);
      Canvas.LineTo(i, Height - 1 - BarLength)
    end;

    Canvas.Pen.Color := clDkGray;
      // Vertical Lines for visual estimation
    for i := 0 to 25 do
    begin
      Canvas.MoveTo(10 * i, Height - 1);
      if i mod 5 = 0
        then Delta := MajorTickSize
      else Delta := MajorTickSize div 2;
      Canvas.LineTo(10 * i, Height - 1 - Delta);
    end;

      // Horizontal Lines
    for j := 0 to 4 do
    begin
      Canvas.MoveTo(0, j * Height div 5);
      Canvas.LineTo(Width - 1, j * Height div 5);
    end;

    Canvas.Brush.Style := bsClear;
    Canvas.TextOut(2, 2, 'Max = ' + IntToStr(MaxValue));
    Canvas.TextOut(2, Height - Canvas.TextHeight('X') -
      MajorTickSize,
      '0');
    Canvas.TextOut(Width - Canvas.TextWidth('250 '),
      Height - Canvas.TextHeight('X') - MajorTickSize, '250')
  end
end {DrawHistogram};


procedure GetHistogram(const Bitmap: TBitmap;
  const ColorPlane: TColorPlane;
  var Histogram: THistogram);
var
  C, M, Y, K: INTEGER; //  CMYK color space
  H, S, V: INTEGER; //  HSV color
  i: INTEGER;
  index: INTEGER;
  j: INTEGER;
  L: INTEGER;
  Row: pRGBTripleArray;
begin
  if (Bitmap.PixelFormat <> pf24bit)
    then raise EHistogramError.Create('GetHistogram:  ' +
      'Bitmap must be 24-bit.');

  index := 0; // avoid compiler warning about initialization

  Histogram.Clear;

    // Step through each row of image.
  for j := Bitmap.Height - 1 downto 0 do
  begin
    Row := Bitmap.Scanline[j];

    for i := Bitmap.Width - 1 downto 0 do
    begin
      case ColorPlane of
        cpY: index := RGBTripleToY(Row[i]);

        cpRGB,
          cpIntensity: index := RGBTripleToIntensity(Row[i]);

        cpRed: index := Row[i].rgbtRed;
        cpGreen: index := Row[i].rgbtGreen;
        cpBlue: index := Row[i].rgbtBlue;

        cpHueHSV:
          begin
            RGBTripleToHSV(Row[i], H, S, V);

              // Rescale from 0..359 to 0..255
            index := MulDiv(H, 255, 359)
          end;

        cpSaturationHSV:
          begin
            RGBTripleToHSV(Row[i], H, S, V);
            index := S
          end;

        cpValue:
          begin
            RGBTripleToHSV(Row[i], H, S, V);
            index := V
          end;

        cpHueHLS:
          begin
            RGBTripleToHLS(Row[i], H, L, S);
            index := MulDiv(H, 255, 360)
          end;

        cpSaturationHLS:
          begin
            RGBTripleToHLS(Row[i], H, L, S);
            index := S
          end;

        cpCyan:
          begin
            RGBTripleToCMYK(Row[i], C, M, Y, K);
            index := C
          end;

        cpMagenta:
          begin
            RGBTripleToCMYK(Row[i], C, M, Y, K);
            index := M
          end;

        cpYellow:
          begin
            RGBTripleToCMYK(Row[i], C, M, Y, K);
            index := Y
          end;

        cpBlack:
          begin
            RGBTripleToCMYK(Row[i], C, M, Y, K);
            index := K
          end;

        cpLightness:
          begin
            index := RGBTripleToLightness(Row[i])
          end

      end;

        // Debug color conversion problems
      if (index < 0) or (index > 255)
        then begin
        MessageDlgPos('Histogram index = ' + IntToStr(index) +
          ' (i,j) = (' + IntToStr(i) + ', ' +
          IntToStr(j) + ')',
          mtInformation,
          [mbOK],
          0,
          MessageX,
          MessageY);
        index := 0;
      end;

      Histogram.Increment(index)
    end

  end

end {GetHistogram};


  // Single function to create all three R, G and B Histograms for all or part
  // of a bitmap.  Use Bitmap.Canvas.ClipRect as the second parameters to look
  // at the whole bitmap.
  //
  // The calling routine must free the resulting TRGBHistograms object.

function GetRGBHistograms(const Bitmap: TBitmap;
  const Rect: TRect): TRGBHistograms;
var
  i: INTEGER;
  j: INTEGER;
  Row: pRGBTripleArray;
begin
  if Bitmap.PixelFormat <> pf24bit
    then raise EHistogramError.Create('GetRGBHistograms:  ' +
      'Bitmap must be 24-bit.');

  RESULT := TRGBHistograms.Create;

    // Step through each requested row of image.
  for j := Rect.Top to Rect.Bottom - 1 do
  begin
    Row := Bitmap.Scanline[j];

      // Look at each requested pixel within a row and increment histogram stats.
    for i := Rect.Left to Rect.Right - 1 do
    begin
      RESULT.Increment(Row[i])
    end

  end

end {GetRGBHistograms};


procedure GetBitmapStatistics(const Bitmap: TBitmap;
  const ColorPlane: TColorPlane;
  var Statistics: TDescriptiveStatistics);
var
  C, M, Y, K: INTEGER;
  H, S, V: INTEGER; // color coordinates
  i: INTEGER;
  j: INTEGER;
  L: INTEGER;
  Row: pRGBTripleArray;
  Value: INTEGER;
begin
  if Bitmap.PixelFormat <> pf24bit
    then raise EHistogramError.Create('GetBitmapStatistics:  ' +
      'Bitmap must be 24-bit color.');

    // Step through each row of image.
  for j := Bitmap.Height - 1 downto 0 do
  begin
    Row := Bitmap.Scanline[j];

    for i := Bitmap.Width - 1 downto 0 do
    begin
      case ColorPlane of
        cpRGB,
          cpIntensity: value := RGBTripleToIntensity(Row[i]);

        cpRed: value := Row[i].rgbtRed;
        cpGreen: value := Row[i].rgbtGreen;
        cpBlue: value := Row[i].rgbtBlue;

        cpHueHSV:
          begin
            RGBTripleToHSV(Row[i], H, S, V);
            value := H
          end;

        cpSaturationHSV:
          begin
            RGBTripleToHSV(Row[i], H, S, V);
            value := S
          end;

        cpValue:
          begin
            RGBTripleToHSV(Row[i], H, S, V);
            value := V
          end;

        cpHueHLS:
          begin
            RGBTripleToHLS(Row[i], H, L, S);
            value := H
          end;

        cpSaturationHLS:
          begin
            RGBTripleToHLS(Row[i], H, L, S);
            value := S
          end;

        cpLightness:
          begin
            RGBTripleToHLS(Row[i], H, L, S);
            value := L
          end;

        cpCyan:
          begin
            RGBTripleToCMYK(Row[i], C, M, Y, K);
            value := C
          end;

        cpMagenta:
          begin
            RGBTripleToCMYK(Row[i], C, M, Y, K);
            value := M
          end;

        cpYellow:
          begin
            RGBTripleToCMYK(Row[i], C, M, Y, K);
            value := Y
          end;

        cpBlack:
          begin
            RGBTripleToCMYK(Row[i], C, M, Y, K);
            value := K
          end;

      else
        value := 0
      end;

      Statistics.NextValue(value)

    end

  end

end {GetBitmapStatistics};


end.
