//
//  Statistics Library
//  Earl F. Glynn, August 1997
//  Parabolic fit added April 1998
//


unit gUStatisticsLibrary;

interface

uses
  Math; {MaxDouble}

const
  NAN = MaxDouble / 2.0; {Get "Not a Number" properly defined someday}

type

   {===================================================================}

  TDescriptiveStatistics =
    class(TObject)
  protected
    FCount: INTEGER;

    FMaxValue: DOUBLE;
    FMinValue: DOUBLE;

    FSum: DOUBLE;
    FSumOfSquares: DOUBLE;

  public

    property Count: INTEGER read FCount;
    property MaxValue: DOUBLE read FMaxValue;
    property MinValue: DOUBLE read FMinValue;
    property Sum: DOUBLE read FSum;

    constructor Create;

    procedure NextValue(const x: DOUBLE);

    function MeanValue: DOUBLE;
    function StandardDeviation: DOUBLE;

    procedure ResetValues;

  end;


  {===================================================================}

   // Formulas from HP-45 Applications Book, pp. 72, 77, 1974
  TLinearRegression =
    class(TObject)
  protected
    FCount: INTEGER;

    FSumX: DOUBLE;
    FSumXSquared: DOUBLE;

    FSumY: DOUBLE;
    FSumYSquared: DOUBLE;

    FSumXY: DOUBLE;

  public
    constructor Create;
    function CorrelationCoefficient: DOUBLE;
    procedure GetLineCoefficients(var A, B: DOUBLE);
    function RSquared: DOUBLE;
    procedure NextValue(const x, y: DOUBLE);
    procedure ResetValues;

  end;


   {===================================================================}

  TParabolicFit =
    class(TObject)
  protected
    FCount: INTEGER;
    FSumX: array[1..4] of DOUBLE;
    FSumXY: array[1..3] of DOUBLE;

  public
    constructor Create;
    procedure GetCoefficients(var A, B, C: DOUBLE);
    procedure NextValue(const x, y: DOUBLE);
    procedure ResetValues;
  end;

   {===================================================================}

function MedianInteger(x: array of INTEGER): INTEGER;


implementation

uses
  SysUtils; // EInvalidOp


{=====================================================================}


  // 2-by-2 determinant

function Det2(const a, b, c, d: DOUBLE): DOUBLE;
begin
  RESULT := a * d - b * c
end {Det2};


function Det3(const a, b, c, d, e, f, g, h, i: DOUBLE): DOUBLE;
begin
  RESULT := a * Det2(e, f, h, i) - b * Det2(d, f, g, i) + c * Det2(d,
    e, g, h)
end {Det3};


{==== TDescriptiveStatistics =========================================}

constructor TDescriptiveStatistics.Create;
begin
  ResetValues
end {Create};


procedure TDescriptiveStatistics.NextValue(const x: DOUBLE);
begin
  INC(FCount);
  FSum := FSum + x;
  FSumOfSquares := FSumOfSquares + x * x;

  if x > FMaxValue
    then FmaxValue := x;

  if x < FMinValue
    then FMinValue := x
end {NextNumber};


function TDescriptiveStatistics.MeanValue: DOUBLE;
begin
  if FCount = 0
    then RESULT := NAN
  else RESULT := FSum / FCount
end {MeanValue};


function TDescriptiveStatistics.StandardDeviation: DOUBLE;
begin
  if FCount < 2
    then RESULT := NAN
  else RESULT := SQRT((FSumOfSquares - (FSum * FSum / FCount)) /
      (FCount - 1))
end {StandardDeviation};


procedure TDescriptiveStatistics.ResetValues;
begin
  FCount := 0;

  FMinValue := MaxDouble;
  FMaxValue := -MaxDouble;

  FSum := 0.0;
  FSumOfSquares := 0.0
end {ResetValues};

{==== TLinearRegression ==============================================}

constructor TLinearRegression.Create;
begin
  ResetValues
end {Create};


  {Formulas from HP-45 Applications Book, p. 77, 1974}

procedure TLinearRegression.GetLineCoefficients(var A, B: DOUBLE);
begin
  if FCount < 1
    then begin
    A := NAN;
    B := NAN
  end
  else begin
    A := (FSumXY - FSumX * FSumY / FCount) /
      (FSumXSquared - FSumX * FSumX / FCount);
    B := (FSumY - A * FSumX) / FCount
  end
end {GetLineCoefficients};


 {Formula from HP-45 Applications Book, p. 72, 1974}

function TLinearRegression.CorrelationCoefficient: DOUBLE;
var
  numerator: DOUBLE;
  denominator: DOUBLE;
begin
  if FCount < 2
    then RESULT := NAN
  else begin
    numerator := FSumXY - FSumX * FSumY / FCount;
    denominator := SQRT((FSumXSquared - FSumX * FSumX / FCount) *
      (FSumYSquared - FSumY * FSumY / FCount));
    try
      RESULT := numerator / denominator
    except
      on EInvalidOp
        do if (numerator = 0.0) and (denominator = 0.0)
        then RESULT := 1.0 // all zeros have "perfect" correlation
        else RESULT := NAN;
    end
  end
end {CorrelationCoefficient};


function TLinearRegression.RSquared: DOUBLE;
begin
  RESULT := 0.0 {FIX THIS}
end {RSquared};


procedure TLinearRegression.NextValue(const x, y: DOUBLE);
begin
  INC(FCount);

  FSumX := FSumX + x;
  FSumXSquared := FSumXSquared + x * x;

  FSumY := FSumY + y;
  FSumYSquared := FSumYSquared + y * y;

  FSumXY := FSumXY + x * y
end {NextValue};


procedure TLinearRegression.ResetValues;
begin
  FCount := 0;

  FSumX := 0.0;
  FSumXSquared := 0.0;

  FSumY := 0.0;
  FSumYSquared := 0.0;

  FSumXY := 0.0
end {ResetValues};

{==== TParabolicFit ==================================================}

  // Use "brute force" for now

constructor TParabolicFit.Create;
begin
  ResetValues
end {Create};


procedure TParabolicFit.GetCoefficients(var A, B, C: DOUBLE);
var
  denominator: DOUBLE;
begin
  denominator := Det3(FCount, FSumX[1], FSumX[2],
    FSumX[1], FSumX[2], FSumX[3],
    FSumX[2], FSumX[3], FSumX[4]);
  if (FCount < 3) or (ABS(denominator - 1E-5) <= 1E-5)
    then begin
    A := NAN;
    B := NAN;
    C := NAN;
  end
  else begin
    A := Det3(FSumXY[1], FSumX[1], FSumX[2],
      FSumXY[2], FSumX[2], FSumX[3],
      FSumXY[3], FSumX[3], FSumX[4]) / denominator;

    B := Det3(FCount, FSumXY[1], FSumX[2],
      FSumX[1], FSumXY[2], FSumX[3],
      FSumX[2], FSumXY[3], FSumX[4]) / denominator;

    C := Det3(FCount, FSumX[1], FSumXY[1],
      FSumX[1], FSumX[2], FSumXY[2],
      FSumX[2], FSumX[3], FSumXY[3]) / denominator;
  end
end {GetCoefficients};


procedure TParabolicFit.NextValue(const x, y: DOUBLE);
var
  Term: DOUBLE;
begin
  INC(FCount);

  FSumX[1] := FSumX[1] + x;

  Term := x * x;
  FSumX[2] := FSumX[2] + Term;


  Term := Term * x;
  FSumX[3] := FSumX[3] + Term;

  Term := Term * x;
  FSumX[4] := FSumX[4] + Term;

  FSumXY[1] := FSumXY[1] + y;

  Term := x * y;
  FSumXY[2] := FSumXY[2] + Term;

  Term := Term * x;
  FSumXY[3] := FSumXY[3] + Term
end {NextValue};


procedure TParabolicFit.ResetValues;
var
  i: INTEGER;
begin
  FCount := 0;

  for i := Low(FSumx) to High(FSumX) do
    FSumX[i] := 0.0;

  for i := Low(FSumXY) to High(FSumXY) do
    FSumXY[i] := 0.0
end {ResetValues};


{==== Median =========================================================}

function MedianInteger(x: array of INTEGER): INTEGER;
var
  i: INTEGER;
  j: INTEGER;
  Middle: INTEGER;
  Temporary: INTEGER;
begin
    {Use truncated selection sort to find median}

  Middle := (High(x) + 1) div 2;

  for i := 0 to Middle do
  begin
    for j := 1 to High(x) - i do
    begin
      if x[j] > x[j - 1]
        then begin
        Temporary := x[j];
        x[j] := x[j - 1];
        x[j - 1] := Temporary
      end
    end

  end;

  if Odd(High(x))
    then begin
      {When High(x) is Odd, there are an even number of elements in array.
       Define median as average of two middle values.}
    RESULT := (x[middle] + x[middle - 1]) div 2
  end
  else begin
       {When High(x) is Even, there are an odd number of elements in array.
       Median is the middle value.}
    RESULT := x[middle]
  end
end {MedianInteger};


{=====================================================================}

end.
